#!/bin/env ruby

module Bias
  # Return a biased random float greater than or equal to 0 and less than 1
  # - If max is supplied, return a biased random integer greater than or equal 
  #   to 0 and less than max.
  # - weight defines the bias:
  #   - A weight of 0 will result in a random distribution
  #   - A weight of -1 will always return 0
  #   - A weight of 1 will always return the maximum value
  #   - Any other weight will bias towars the min or max value with an 
  #     exponential distribution
  def self.biased_rand(weight, max=nil)
    unless -1 <= weight && weight <= 1
      raise ArgumentError.new('Weight must be between -1 and 1') 
    end

    x = case weight
    when 0.0
      Kernel.rand
    when -1.0
      0
    when 1.0
      0.99999999999999999 # almost 1
    when 0.0...1.0
      Kernel.rand ** (1 - weight.abs)
    when -1.0...0.0
      1 - Kernel.rand ** (1 - weight.abs)
    else
      0
    end

    max ? (x * max).to_i : x
  end
  
  def self.interactive2
    w = ARGV[1].to_f
    a = %w{ cow sheep horse pig rabbit hen }
    while a.any?
      puts a.biased_pop(w)
    end
  end

  def self.interactive
    b = ARGV[0].to_i
    w = ARGV[1].to_f
    n = 100_000

    puts "buckets: #{b}"
    puts "weight:  #{w}"
    puts

    raise ArgumentError, "need at least one bucket" unless b > 0

    buckets = Array.new(b, 0)

    n.times do
      x = biased_rand(w, b)
      buckets[x] += 1
    end

    buckets.each_with_index do |count, i|
      puts sprintf("bucket %2d: %6.2f%\n", i, (count*100/n.to_f))
    end
  end
end

# Open the Array class to add a biased_pop method
# This will remove and return a biased random element from the array,
# the position of which will be calculated by the biased_rand method
class Array
  def biased_pop(weight)
    delete_at(Bias.biased_rand(weight, size)) if size > 0
  end
end

if __FILE__ == $0
  Bias.interactive
end