#!/usr/bin/env ruby
require 'rubygems'
require 'camping'
require 'camping/session'
require 'mongrel'
require 'mongrel/camping'
require 'lib/bias'

Camping.goes :PushPop

module PushPop
    include Camping::Session
end

module PushPop::Models
  class User < Base
    has_many :pages, :order => 'created_on desc'
    validates_presence_of :email, :password, :bias
    validates_format_of :email, :with => /\A([^@\s]+)@((?:[-a-z0-9]+\.)+[a-z]{2,})\Z/i
    validates_length_of :email, :maximum => 255
    validates_length_of :password, :maximum => 255
    validates_confirmation_of :password
    validates_inclusion_of :bias, :in => -100..100
    def self.authenticate(email, pass)
      find :first, :conditions => [ 'email = ? AND password = ?', email, pass ]
    end
    def pop_next_page
      pages.biased_pop(bias / 100.0) # bias is -100..100, but we need -1..1
    end
    def to_s
      "User(email: #{email}, added at: #{created_on})"
    end
  end
  
  class Page < Base
    belongs_to :user
    validates_presence_of :uri
    validates_length_of :uri, :maximum => 2000
    def to_s
      "Page(id: #{id}, uri: #{uri}, added at: #{created_on})"
    end
  end

  class CreateTheBasics < V 1.0
    def self.up
      create_table :pushpop_users, :force => true do |t|
        t.column :id,         :integer,   :null => false
        t.column :email,      :string,    :null => false, :limit => 255
        t.column :password,   :string,    :null => false, :limit => 255
        t.column :bias,       :integer,   :null => false
        t.column :created_on, :timestamp
      end
      create_table :pushpop_pages, :force => true do |t|
        t.column :id,         :integer,   :null => false
        t.column :user_id,    :integer,   :null => false
        t.column :uri,        :string,    :null => false, :limit => 2000
        t.column :created_on, :timestamp
      end
    end
    def self.down
      drop_table :pushpop_users
      drop_table :pushpop_pages
    end
  end
end

module PushPop::Controllers
  class Index < R '/'
    def get
      if @user = current_user
        render :home
      else
        @user = Models::User.new
        render :index
      end
    end
  end
  
  class Login < R '/login'
    def get
      @user = Models::User.new
      render :login
    end
    def post
      if log_in(input.email, input.password)
        redirect Index
      else
        @user = Models::User.new :email => input.email
        @error = "Bad username or password"
        render :login
      end
    end
  end

  class Logout < R '/logout'
    def get
      log_out
      redirect Index
    end
  end
  
  class Signup < R '/new'
    def get
      @user = Models::User.new
      render :signup
    end
    def post
      @user = Models::User.new :email => input.email, :password => input.password, 
          :password_confirmation => input.password_confirmation, :bias => 0
      if @user.save
        puts @user
        @state.user_id = @user.id
        redirect Index
      else
        render :signup
      end
    end
  end
  
  class Bias < R '/bias'
    def post
      ensure_logged_in do
        @user.bias = input.bias.to_i
        if @user.save
          curr_bias(@user)
        else
          'Error saving new bias value'
        end
      end
    end
  end

  class Push < R '/push', '/(api)/push'
    def get(api=nil)
      ensure_logged_in do
        render :push
      end
    end
    def post(api=nil)
      format = api.blank? ? :html : :json
      ensure_logged_in(format) do
        @uri = input.uri.to_s.strip
        @page = Models::Page.new :user => @user, :uri => @uri
        if @page.save
          case format
          when :json
            json :success
          else
            redirect Index
          end
        else
          case format
          when :json
            json :error, :message => @page.errors.full_messages.join("\n")
          else
            render :home
          end
        end
      end
    end
  end

  class Pop < R '/pop', '/pop/(\d+)', '/(api)/pop', '/(api)/pop/(\d+)'
    def post(*args)
      format = args.include?('api') ? :json : :html
      page = args.reject {|a| a == 'api' }.first
      ensure_logged_in(format) do
        @page = @user.pages.find(page) if page
        @page = @user.pop_next_page unless @page
        if @page
          @uri = @page.uri
          @page.destroy
          case format
          when :json
            json :success, :uri => @uri
          else
            static_redirect @uri
          end
        else
          case format
          when :json
            json :success, :uri => uri_for(Index)
          else
            redirect Index
          end
        end
      end
    end
  end
  
  class Delete < R '/delete/(\d+)'
    def post(page)
      ensure_logged_in(:html) do
        @user.pages.destroy(page)
        redirect Index
      end
    end
  end
  
  class Firefox < R '/firefox'
    def get
      render :firefox
    end
  end

  class Static < R '/static/(.+)'
    MIME_TYPES = {
      '.css' => 'text/css', 
      '.js' => 'text/javascript', 
      '.jpg' => 'image/jpeg',
      '.png' => 'image/png' }
    PATH = File.expand_path(File.dirname(__FILE__))

    def get(path)
      @headers['Content-Type'] = MIME_TYPES[path[/\.\w+$/, 0]] || "text/plain"
      unless path.include? ".." # prevent directory traversal attacks
        @headers['X-Sendfile'] = "#{PATH}/static/#{path}"
      else
        @status = "403"
        "403 - Invalid path"
      end
    end
  end
end

module PushPop::Helpers
  def log_in(email, pass)
    @user = PushPop::Models::User.authenticate email, pass
    @state.user_id = @user.id if @user
    @user
  end
  
  def log_out
    @state.user_id = nil
  end
  
  def current_user
    unless @state.user_id.blank?
      PushPop::Models::User.find @state.user_id
    end
  end
  
  def ensure_logged_in(format=:html)
    if @user = current_user
      yield
    else
      case format
      when :json
        json :login, :uri => uri_for(R(PushPop::Controllers::Login))
      else
        redirect PushPop::Controllers::Login
      end
    end
  end

  def uri_for(local_uri)
    "http:" + URL(local_uri).to_s
  end
  
  def json(status, params={})
    params = { :status => status }.merge params
    '{ ' + params.collect {|k,v| "#{k.to_s}: \"#{v.to_s}\"" }.join(', ') + ' }'
  end
  
  def static_redirect(uri)
    r 302, '', 'Location' => uri
  end
  
  # Stripped down version of ActionView::Helpers::DateHelper#time_ago_in_words
  def time_ago(from_time)
    to_time = Time.now
    distance_in_minutes = (((to_time - from_time).abs)/60).round
    distance_in_seconds = ((to_time - from_time).abs).round

    case distance_in_minutes
      when 0               then "less than a minute"
      when 1               then "1 minute"
      when 2..44           then "#{distance_in_minutes} minutes"
      when 45..89          then 'about 1 hour'
      when 90..1439        then "about #{(distance_in_minutes.to_f / 60.0).round} hours"
      when 1440..2879      then '1 day'
      when 2880..43199     then "#{(distance_in_minutes / 1440).round} days"
      when 43200..86399    then 'about 1 month'
      when 86400..525959   then "#{(distance_in_minutes / 43200).round} months"
      when 525960..1051919 then 'about 1 year'
      else                      "over #{(distance_in_minutes / 525960).round} years"
    end
  end
  
  def bias_str(bias)
    case bias
      when -100     then 'Newest first'
      when -100...0 then "Favour newer pages by #{bias.abs}%"
      when 0        then 'Totally random'
      when 0...100  then "Favour older pages by #{bias}%"
      when 100      then 'Oldest first'
      else               'Error'
    end
  end
  
  def curr_bias(user)
    bias_str(user.bias)
  end
  
  def box(params={})
    div :id => params[:id], :class => 'box' do
      h3 params[:title]
      div :class => 'inner' do
        self << yield
      end
    end
  end
  
  def pushpop_link; a("PushPop", :href => R(PushPop::Controllers::Index)); end
  def firefox_link; a("Firefox", :href => "http://www.mozilla.com/firefox/"); end
end

module PushPop::Views
  def layout
    html do
      head do
        title { 'PushPop' }
        link :rel => 'stylesheet', :href => R(Static, 'screen.css'), :type => 'text/css', :media => 'screen, projection'
        link :rel => 'stylesheet', :href => R(Static, 'print.css'), :type => 'text/css', :media => 'print'
        script :type => 'text/javascript', :src => R(Static, 'custom.js')
      end
      body do
        div :id => 'wrapper' do
          div :id => 'header' do
            h1 { span 'PushPop' }
          end
          div :id => 'content' do
            @sidebar.call if @sidebar
            _branding
            _errors
            self << yield
            _footer
          end
        end
      end
    end
  end

  def index
    login
  end

  def login
    @sidebar = Proc.new { _login }
    _intro_blurb
  end
  
  def signup
    _signup
  end

  def home
    @sidebar = Proc.new { _slider }
    div :class => 'container' do
      _logged_in_blurb
      _push
      _queue
    end
  end
  
  def firefox
    _firefox
  end
  
  # partials
  
  def _errors
    p(:class => 'error') { span "Error: #{@error}" } if @error
  end
  
  def _branding
    h2 do
      a :href => R(Index) do
        text 'PushPop.'
        span ' the web on a stack.'
      end
    end
  end
  
  def _intro_blurb
    div :class => 'unbox' do
      h4 "What?"
      p do
        text "#{ pushpop_link } stores pages to read later. It's not bookmarks - 
          you haven't read these pages yet, so why would you bookmark them? 
          No, the pages stored here are more... #{ em "transient..." }"
      end

      h4 "But... Why?"
      p do
        text "To scratch an itch. Letting the #{ em "tide of the Internets" }
          take you away is great and all, but sometimes one simply isn't in the
          mood to digest that interesting-sounding essay, or doesn't have 
          headphones on to get the full experience of that probably-hilarious 
          youtube clip. So just push it onto your handy #{ pushpop_link } 
          stack and go about your merry business."
      end
      
      h4 "Firefox Love"
      p do
        text "#{ firefox_link } is great. Which is why #{ pushpop_link } has 
        #{ em "full Firefox integration" } in the form of #{ a "fancy-schmancy 
        toolbar buttons", :href => R(Firefox) }."
      end
    end
  end

  def _login
    errors_for @user
    box :id => 'login', :title => 'Log In' do
      form :action => R(Login), :method => 'post' do
        _user_fields
        p { input :type => 'submit', :name => 'login', :value => 'Login' }
      end
      div :class => 'clear'
      p :class => 'nopad' do 
        a "New user?", :href => R(Signup) 
      end
    end
  end
  
  def _signup
    errors_for @user
    box :id => 'signup', :title => 'Create an Account' do
      p "We aren't going to sell your email address or use it for nefarious purposes. We promise."
      form :action => R(Signup), :method => 'post' do
        _user_fields
        p :class => 'field' do
          label 'Password (again)', :for => 'password_confirmation'; br
          input :name => 'password_confirmation', :type => 'password', :value => @user.password_confirmation
        end
        p { input :type => 'submit', :name => 'signup', :value => 'Sign up' }
      end
      p :class => 'nopad' do
        text "Already have an account? "
        a "Log in.", :href => R(Login)
      end
    end
  end
  
  def _user_fields
    p :class => 'field' do
      label 'Email', :for => 'email'; br
      input :name => 'email', :type => 'text', :value => @user.email
    end
    p :class => 'field' do
      label 'Password', :for => 'password'; br
      input :name => 'password', :type => 'password', :value => @user.password
    end
  end

  def _logged_in_blurb
    p do 
      text "You are currently logged in as #{@user.email}. You can #{ a 'log out', 
        :href => R(Logout) } if you like, but you'll need to stay logged in to use 
        the awesome #{ a 'Firefox plugin', :href => R(Firefox) }."
    end
    
    blockquote do
      text "Tip: You can adjust the behaviour of your stack by moving the slider on the right.
        If you move the slider all the way to the top, Pop will always give you the newest 
        page. All the way to the bottom, and Pop will always give you the oldest page. 
        Anywhere in between will add some randomness to the mix, and you can favour newer 
        or older pages as much as you like."
    end
  end
  
  def _push
    box :id => 'push', :title => 'Push a Page' do
      form :action => R(Push), :method => 'post' do
        input :name => 'uri', :type => 'text', :size => '27', :value => @uri
        button 'Push', :type => 'submit', :class => 'positive'
      end
      div :class => 'clear'
    end
  end

  def _queue
    box :id => 'queue', :title => 'The Stack' do
      if @user.pages.any?
        ul do
          @user.pages.each do |page|
            li do
              span :class => 'page_link' do
                a page.uri, :href => page.uri
                text " (added #{time_ago(page.created_on)} ago)"
              end
              a(:href => page.uri, :onclick => "PushPop.post(this, '#{R(Delete, page)}')", :class => 'pop', :title => 'Pop', :target => '_blank') { span 'Pop' }
              a(:href => R(Delete, page), :onclick => 'PushPop.post(this);return false', :class => 'delete', :title => 'Delete') { span 'Delete' }
              div :class => 'clear'
            end
          end
        end
        form :id => 'pop', :action => R(Pop), :method => 'post' do
          button 'Pop', :type => 'submit', :class => 'positive'
        end
        div :class => 'clear'
      else
        p 'No pages yet.'
      end
    end
  end
  
  def _slider
    box :id => 'slider', :title => 'Pop Priority' do
      div(:id => 'track') { div :id => 'handle' }
      dl do
        dt 'Slider at top:'; dd 'Newest pages first'
        dt 'Slider at middle:'; dd 'Totally random'
        dt 'Slider at bottom:'; dd 'Oldest pages first'
        dt 'Current setting:', :class => 'current'
        dd :class => 'current' do
          img :id => 'biasSpinner', :src => R(Static, 'spinner.gif'), :style => 'display:none'
          span curr_bias(@user), :id => 'bias'
        end
      end
      script :type => 'text/javascript', :src => R(Static, 'lib/prototype.js')
      script :type => 'text/javascript', :src => R(Static, 'lib/slider.js')
      script "PushPop.initSlider('#{R(Bias)}',#{@user.bias});", :type => 'text/javascript'
    end
  end
  
  def _footer
    div :id => 'footer' do
      p do
        a 'PushPop', :href => R(Index)
        text ' was brought to you by '
        a 'Ken Pratt', :href => 'http://kenpratt.net/'
        text ' in the fine year of '
        a '2007', :href => 'http://en.wikipedia.org/wiki/2007'
        text '. Enjoy.'
      end
    end
  end
  
  def _firefox
    div :class => 'unbox' do
      script :type => 'application/x-javascript', :src => R(Static, 'extension.js')
      p :id => 'install_button' do
        a "Install PushPop Firefox Extension", 
          :href => R(Static, 'pushpop.xpi'), 
          :onclick => "return install(event, 'PushPop', '#{R(Static, 'logo-square.png')}');"
      end
      
      h4 "Firefox Integration"
      p do
        text "#{ pushpop_link } is pretty useful on its own, but it really shines when used 
          with the #{ em "PushPop Firefox Extension" }. The extension adds two new buttons
          to the Firefox toolbar:"
      end

      h5 { text "Push "; img :src => R(Static, 'push.png') }
      p do 
        text "The #{ em "Push" } button adds the current page to your PushPop stack, 
          and closes the current tab." 
      end

      h5 { text "Pop "; img :src => R(Static, 'pop.png') }
      p do 
        text "The #{ em "Pop" } button opens a page from your PushPop stack in
          a new tab, removing that page from your stack. If the page wasn't what you were
          looking for, simply #{ em "Push" } it back onto your stack."
      end

      #h4 "Demo"
      #p { img :src => R(Static, 'push.png'); img :src => R(Static, 'pop.png') }
      #p { img :src => R(Static, 'firefox-buttons.png') }
      #p { img :src => R(Static, 'firefox-pop1.png') }
      #p { img :src => R(Static, 'firefox-pop2.png') }
    end
  end
end

def PushPop.create
  Camping::Models::Session.create_schema
  PushPop::Models.create_schema :assume => (PushPop::Models::Page.table_exists? ? 1.0 : 0.0)
end

if __FILE__ == $0
  PushPop::Models::Base.establish_connection :adapter => 'sqlite3', :database => 'pushpop.db'
  PushPop::Models::Base.logger = Logger.new('camping.log')
  PushPop::Models::Base.threaded_connections = false
  PushPop.create
  
  h = Mongrel::HttpServer.new('0.0.0.0', 3000)
  h.register('/', Mongrel::Camping::CampingHandler.new(PushPop))
  h.register('/favicon.ico', Mongrel::Error404Handler.new(''))
  puts '** PushPop is now running at http://127.0.0.1:3000/'
  h.run.join
end
