require 'rubygems'
require 'mosquito'
require File.dirname(__FILE__) + '/../pushpop'

PushPop.create
include PushPop::Models

def all_fixtures
  fixtures :pushpop_users, :pushpop_pages
end

class Camping::FunctionalTest
  def local_uri_regex(path="")
    "http://#{@request['SERVER_NAME']}/(#{@class_name_abbr.downcase}/)?#{path}"
  end

  def login(user=:buster)
    post '/login', :email => "#{user}@bluth.com", :password => 'test'
  end

  def assert_static_redirect(uri)
    assert_equal uri, @response.headers['Location']
  end

  def clever_assert_response(status_code)
    puts @response.to_s.gsub(/(\<\/.+?\>)/, '\1'+"\n") if @response.status >= 500
    dumb_assert_response status_code
  end

  alias_method :dumb_assert_response, :assert_response
  alias_method :assert_response, :clever_assert_response
end

class Camping::UnitTest
end

class TestUsers < Camping::UnitTest
  all_fixtures

  def test_nothing
    assert true
  end
  
  def test_pop_with_top_bias
    user = User.find 5
    assert_equal 'http://one.ca', user.pop_next_page.uri
    assert_equal 'http://two.ca', user.pop_next_page.uri
    assert_equal 'http://three.ca', user.pop_next_page.uri
    assert_equal 'http://four.ca', user.pop_next_page.uri
    assert_nil user.pop_next_page
  end

  def test_pop_with_bottom_bias
    user = User.find 6
    assert_equal 'http://four.ca', user.pop_next_page.uri
    assert_equal 'http://three.ca', user.pop_next_page.uri
    assert_equal 'http://two.ca', user.pop_next_page.uri
    assert_equal 'http://one.ca', user.pop_next_page.uri
    assert_nil user.pop_next_page
  end
end

class TestPushPop < Camping::FunctionalTest
  all_fixtures
    
  def test_login
    post '/login', :email => 'buster@bluth.com', :password => 'test'
    assert_response :redirect
    assert_redirected_to '/'
  end
  
  def test_update_bias
    login
    post '/bias', :bias => 100
    assert_response :success
    assert_match_body 'Oldest first'
  end
  
  def test_update_bias_bad_value
    login
    post '/bias', :bias => 200
    assert_response :success
    assert_match_body 'Error saving new bias value'
  end
  
  def test_regular_push_not_logged_in
    post '/push', :uri => 'http://google.ca/'
    assert_response :redirect
    assert_redirected_to '/login'
  end

  def test_regular_push
    login
    post '/push', :uri => 'http://google.ca/'
    assert_response :redirect
    assert_redirected_to '/'
  end

  def test_regular_push_bad_uri
    login
    post '/push'
    assert_response :success
    assert_match_body /action=".*\/push"/
  end
  
  def test_regular_pop_not_logged_in
    post '/pop'
    assert_response :redirect
    assert_redirected_to '/login'
  end

  def test_regular_pop
    login :gob
    post '/pop'
    assert_response :redirect
    assert_static_redirect 'http://something.com/'
  end

  def test_regular_pop_empty_queue
    login :michael
    post '/pop'
    assert_response :redirect
    assert_redirected_to '/'
  end
  
  def test_pop_specific_page
    login :maeby
    post '/pop/4'
    assert_response :redirect
    assert_static_redirect 'http://four.ca/'
  end
  
  def test_delete
    login :maeby
    post '/delete/5'
    assert_response :redirect
    assert_redirected_to '/'
  end
  
  def test_api_push_not_logged_in
    post '/api/push', :uri => 'http://google.ca/'
    assert_response :success
    assert_match_body 'status: "login"'
    assert_match_body /uri: "#{local_uri_regex('login')}"/
  end

  def test_api_push
    login
    post '/api/push', :uri => 'http://google.ca/'
    assert_response :success
    assert_match_body 'status: "success"'
  end

  def test_api_push_bad_uri
    login
    post '/api/push'
    assert_response :success
    assert_match_body 'status: "error"'
    assert_match_body /message: ".+"/
  end
  
  def test_api_pop_not_logged_in
    post '/api/pop'
    assert_response :success
    assert_match_body 'status: "login"'
    assert_match_body /uri: "#{local_uri_regex('login')}"/
  end

  def test_api_pop
    login :gob
    post '/api/pop'
    assert_response :success
    assert_match_body 'status: "success"'
    assert_match_body 'uri: "http://something.com/"'
  end

  def test_api_pop_empty_queue
    login :michael
    post '/api/pop'
    assert_response :success
    assert_match_body 'status: "success"'
    assert_match_body /uri: "#{local_uri_regex}"/
  end
end
